$Title OptORF example - microbial strain design

$Ontext

   The objective of this model is to identify metabolic engineering strategies
   to maximize the biochemical production at the optimal growth condition.

   CPLEX is used to solve the problem using the indicator constraint facility.
   The indicator constraints should be re-formulated using the big-M method
   in order to be solved by other MIP solvers.
   
   Users may need to adjust the tolerances and optimality criteria
   when applying to large networks.

   The program will generate a text file called OptORF_example.txt that 
   contains the strain design strategies.

Joonhoon Kim and Jennifer L. Reed
Department of Chemical Engineering,
University of Wisconsin-Madison, Madison, Wisconsin
March 1, 2010

$Offtext


Sets
soln             number of solutions     /s1*s4/
i                metabolites             /Substrate,I1,I2,P1,P2,Biomass/
j                reactions               /v1,v2,v3,v4,v5,b1,b2,b3,b4/
rev(j)           reversible reactions    /v3,b1/
bnd(j)           constrained reactions   /b1,b4/
g                genes                   /G1A,G1B,G2,G3,G4,G5,G6,TF1/
m                conditions              /1/
n                enzymes                 /E1,E2,E3,E4,E5,E6/
r                regulators              /TF1,Substrate/
TF(r)            transcription factors   /TF1/
med(r)           medium conditions       /Substrate/
Act(g,m,r)       activators              /G3.1.TF1,G5.1.TF1,TF1.1.Substrate/
Rep(g,m,r)       repressors              /G1A.1.TF1/
Rmed(j,med)      reaction to medium      /b1.Substrate/
GPR(j,n,g)       GPR associations        /v1.E1.G1A,v1.E1.G1B,v2.E2.G2,v3.E3.G3,
                                          v4.E4.G4,v5.E5.G5,v5.E6.G6/
rgm(g,m)         gene to condition
rg(g)            regulated genes
md(j)            regulator reaction
rr(j,n)          protein to reaction
jk(j)            reactions with GPR
jl(j)            reactions without GPR;

Parameters
S(i,j)           Stoichiometry           /Substrate.v1 -1, I1.v1 1,
                                          I1.v2 -1, P1.v2 1, Biomass.v2 0.08,
                                          Substrate.v3 -1, I2.v3 1,
                                          I1.v4 -1, I2.v4 1,
                                          I2.v5 -1, P2.v5 1, Biomass.v5 0.12,
                                          Substrate.b1 -1, P1.b2 -1,
                                          P2.b3 -1, Biomass.b4 -1/
LowerLimits(j)   Lower bounds for rxns;

rgm(g,m)$sum(r$(Act(g,m,r) or Rep(g,m,r)),1)=yes;
rg(g)$sum(m$rgm(g,m),1)=yes;
md(j)$sum(med$Rmed(j,med),1)=yes;
rr(j,n)$sum(g$GPR(j,n,g),1)=yes;
jk(j)$sum(n$rr(j,n),1)=yes;
jl(j)=not jk(j);

* Set bounds for reversible and irreversible reactions
LowerLimits(rev)=-inf;
LowerLimits(j)$(not rev(j))=0;

* Set substrate uptake rate and minimum growth rate
LowerLimits('b1')=-10;
LowerLimits('b4')=0.1;

Scalar
alpha            minimum increase of the objective per deletion
Zmax             maximum number of knockouts
Wmax             maximum number of overexpression
epsilon          threshold for an active flux /1e-3/
iter;

Parameters
c(j)             used to define the objective function for outer problem
pri(j)           used to define the objective function for inner problem
storeZ(g,soln)   store knockout states for each solution
storeW(g,soln)   store overexpression states for each solution
storeKnock(soln) store number of gene knockouts for each solution
storeOverx(soln) store number of gene overexpression for each solution
storeV(j,soln)   store reaction flux value for each solution
maxobj(soln)     maximum objective value for each solution
minobj(soln)     minimum objective value for each solution;

Variables
v(j)             flux values through reaction in network
u1(i)            dual variable for S*v = 0
u2(j)            dual variable for   v = 0
Obj              the objective function;

Positive Variables
u3(j)            dual variable for -v =l= -vmin;

Binary Variables
x(r)             indicator for transcriptional regulators
y(g)             indicator for genes
yy(g)            indicator for expression
w(g)             indicator for overexpression
z(g)             indicator for knockout
a(g,m)           indicator for regulatory rules
b(j,n)           indicator for enzyme complexes
d(j)             indicator for reactions;

Alias(p,g);

Equation
optobj           OptORF objective function
fbaobj           FBA objective function
maxknock         maximum number of allowed knockouts
maxoverx         maximum number of allowed overexpression
primaldual       primal objective = dual objective
massbalance(i)   steady state mass balance
dualbalance(j)   dual constraint for primal variable v
vbound(j)        set bound for v from reaction indicators
u2bound(j)       set bound for u2 from reaction indicators
u3bound(j)       set bound for u3 from reaction indicators
mdrulesup(r)     determine what is present in environment
mdruleslo(r)     determine what is present in environment
grules1(g,m)     determine regulatory rules from transcriptional regulators
grules2(g,m,r)   determine regulatory rules from transcriptional regulators
grules3(g,m,r)   determine regulatory rules from transcriptional regulators
grules4(g)       determine gene states from regulatory rules
grules5(g,m)     determine gene states from regulatory rules
perturb1(g)      determine gene knockout or overexpression
perturb2(g)      determine gene knockout or overexpression
prules1(j,n)     determine protein states from regulatory rules
prules2(j,n,g)   determine protein states from regulatory rules
prules3(j)       determine reaction states from enzyme comples states
prules4(j,n)     determine reaction states from enzyme comples states
intcut(soln)     integer cut to generate diverse solutions;

optobj..         Obj=e=sum(j, c(j)*v(j))-alpha*sum(g, w(g)+z(g));
fbaobj..         Obj=e=sum(j, c(j)*v(j));

maxknock..       sum(g, z(g))=l=Zmax;
maxoverx..       sum(g, w(g))=l=Wmax;

primaldual..     sum(j, pri(j)*v(j))=e=-sum( j$bnd(j),v.lo(j)*u3(j) );
massbalance(i).. sum(j, S(i,j)*v(j))=e=0;
dualbalance(j).. sum(i, S(i,j)*u1(i))+u2(j)-u3(j)$bnd(j)=e=pri(j);

vbound(jk)..     v(jk)=e=0;
u2bound(jk)..    u2(jk)=e=0;
u3bound(jk)..    u3(jk)=e=0;

perturb1(g)..    w(g)-z(g)=e=yy(g)-y(g);
perturb2(g)..    w(g)+z(g)=l=1;

mdrulesup(med).. sum(md$Rmed(md,med), v(md))=g=-epsilon;
mdruleslo(med).. sum(md$Rmed(md,med), v(md))=l=-epsilon;

grules1(rgm(g,m))..   a(g,m)-1=g=sum(r$Act(g,m,r), (x(r)$(not TF(r))+sum(p$sameas(p,r), yy(p))$TF(r))-1)
                                -sum(r$Rep(g,m,r), (x(r)$(not TF(r))+sum(p$sameas(p,r), yy(p))$TF(r)));
grules2(Act(g,m,r)).. a(g,m)=l=(x(r)$(not TF(r))+sum(p$sameas(p,r), yy(p))$TF(r));
grules3(Rep(g,m,r)).. a(g,m)=l=1-(x(r)$(not TF(r))+sum(p$sameas(p,r), yy(p))$TF(r));
grules4(rg(g))..      y(g)=l=sum(m$rgm(g,m), a(g,m));
grules5(rgm(g,m))..   y(g)=g=a(g,m);

prules1(rr(j,n))..    sum(g$GPR(j,n,g), yy(g)-1)=l=b(j,n)-1;
prules2(GPR(j,n,g)).. yy(g)=g=b(j,n);
prules3(jk)..         d(jk)=l=sum(n$rr(jk,n), b(jk,n));
prules4(rr(jk,n))..   d(jk)=g=b(jk,n);

intcut(soln)..   sum(g,storeZ(g,soln)*z(g))+sum(g,storeW(g,soln)*w(g))+1
                 =l=storeKnock(soln)+storeOverx(soln);

Model FBA    /fbaobj,massbalance/;
Model SrFBA  /fbaobj,massbalance,vbound,
              perturb1,perturb2,mdrulesup,mdruleslo,
              grules1,grules2,grules3,grules4,grules5,
              prules1,prules2,prules3,prules4/;
Model OptORF /optobj,primaldual,massbalance,dualbalance,
              vbound,u2bound,u3bound,maxknock,maxoverx,intcut,
              perturb1,perturb2,mdrulesup,mdruleslo,
              grules1,grules2,grules3,grules4,grules5,
              prules1,prules2,prules3,prules4/;

* Use option files and set relative gap to zero
SrFBA.optfile=1;
SrFBA.optcr=0;
OptORF.optfile=2;
OptORF.optcr=0;

* Assign lower and upper bounds for fluxes
v.lo(j)=LowerLimits(j);
v.up(j)=inf;
* Genes without regulation are expressed
y.fx(g)$(not rg(g))=1;
* Reactions without GPR associations are present
d.fx(jl)=1;u2.fx(jl)=0;

* Maximize wildtype biomass production
c('b4')=1;
* Metabolic model
solve FBA using lp maximizing Obj;
* Metabolic and regulatory model
* No gene deletion or overexpression for wildtype
z.fx(g)=0;w.fx(g)=0;
solve SrFBA using mip maximizing Obj;

* Define the outer and inner objectives
c('b4')=0;
c('b2')=1;
pri('b4')=1;

* Calculate theoretical maximum using metabolic model
solve FBA using lp maximizing Obj;
* Set the minimum increase of the objective per deletion
abort$(Obj.l=0) "Maximum production rate is zero";
alpha=1E-3;

* Define target genes for deletion and overexpression
z.lo(g)=0;z.up(g)=1;
w.lo(g)=0;w.up(g)=1;
* Prevent artificial activation of TFs
w.fx(g)$sum(TF$sameas(g,TF),1)=0;

* Define the allowed number of knockouts
Zmax=2;
Wmax=2;

* Initialize parameters for storage
storeZ(g,soln)=0;        StoreW(g,soln)=0;       storeV(j,soln)=0;
storeKnock(soln)=Zmax;   storeOverx(soln)=Wmax;
maxobj(soln)=0;          minobj(soln)=0;

Alias(soln,temp);

loop(temp,
* Solve OptORF
         solve OptORF using mip maximizing Obj;
* Store results
         storeZ(g,temp)=z.l(g);
         storeW(g,temp)=w.l(g);
         storeKnock(temp)=sum(g, z.l(g));
         storeOverx(temp)=sum(g, w.l(g));
         storeV(j,temp)=v.l(j);
         maxobj(temp)=sum(j$c(j), v.l(j));
* Calculate the variability
         z.fx(g)=StoreZ(g,temp);
         w.fx(g)=storeW(g,temp);
         v.fx(j)$pri(j)=storeV(j,temp);
         solve SrFBA using mip minimizing Obj;
         minobj(temp)=obj.l;
* Initialize
         v.lo(j)=LowerLimits(j);
         v.up(j)=inf;
         z.lo(g)=0;z.up(g)=1;
         w.lo(g)=0;w.up(g)=1;
         w.fx(g)$sum(TF$sameas(g,TF),1)=0;
);

*This section writes out the results to a txt file
file result /OptORF_example.txt/;
result.pw=1000; result.pc=5; result.ps=130;
put result;
put 'Knockout'; put "Genes";
iter=1; while(iter lt Zmax, put ''; iter=iter+1;);
put 'Overexpress'; put "Genes";
iter=1; while(iter lt Wmax, put ''; iter=iter+1;);
put 'Growth'; put 'MaxProd'; put 'MinProd'; put /;
loop(temp,
         iter=1;
         put storeKnock(temp);
         loop(g$(storeZ(g,temp) = 1), put g.tl; );
         iter=storeKnock(temp);
         while(iter lt Zmax, put ''; iter=iter+1;);
         put storeOverx(temp);
         loop(g$(storeW(g,temp) = 1), put g.tl; );
         iter=storeOverx(temp);
         while(iter lt Wmax, put ''; iter=iter+1;);
         put storeV('b4',temp):8:4;
         put maxobj(temp):8:4;
         put minobj(temp):8:4;
         put /;
);
putclose result;

* CPLEX option files required for SR-FBA and OptORF

$onecho > cplex.opt
eprhs 1e-9
epopt 1e-9
epint 0
indic vbound(jk)$d(jk) 0
indic mdrulesup(md)$x(md) 0
indic mdruleslo(md)$x(md) 1
$offecho

$onecho > cplex.op2
indic vbound(jk)$d(jk) 0
indic u2bound(jk)$d(jk) 1
indic u3bound(jk)$d(jk) 0
indic mdrulesup(md)$x(md) 0
indic mdruleslo(md)$x(md) 1
$offecho
